#ifndef _MD_FILE_H
#define _MD_FILE_H

/*
  This header file is for generic file tasks such as copying or checking
  the existence or type of a given file or path.
*/

#include <algorithm>
#include <string>
#include <sys/stat.h>

#ifdef _WIN32
//  Windows Specific  - CopyFile()
#include <Windows.h>
#include <direct.h>
#include "dirent.h"
#elif __APPLE__
//  OSX Specific      - copyfile()
#include <copyfile.h>
#include <dirent.h>
#elif __linux
//  Linux Specific    - sendfile(), open(), fstat()
#include <sys/types.h>
#include <fcntl.h>
#include <dirent.h>
#endif

namespace File
{
  struct MatchPathSeparator
  {
    bool operator()(char ch) const
    {
      return ch == '\\' || ch == '/';
    }
  };

  /*
  Determines whether or not a given path is a directory.

  @return True if path is a directory.
  */
  inline bool is_dir(std::string path)
  {
    struct stat buffer;
    return (stat(path.c_str(), &buffer) == 0 && (buffer.st_mode & S_IFDIR));
  }

  /*
  Determines whether or not a given path is a file.

  @return True if path is a file.
  */
  inline bool is_file(std::string path)
  {
    struct stat buffer;
    return (stat(path.c_str(), &buffer) == 0 && (buffer.st_mode & S_IFREG));
  }

  /*
    Returns the file extension for the given path. Potentially unsafe if not passed a correct name.

    @return String with anything past the last .
  */
  inline std::string extension(std::string const& filename)
  {
    return is_file(filename) ? filename.substr(filename.find_last_of(".") + 1) : "";
  }

  inline std::string basename(std::string const& pathname)
  {
    return std::string(std::find_if(pathname.rbegin(), pathname.rend(), MatchPathSeparator()).base(), pathname.end());
  }


  inline void copy(std::string infile, std::string outfile)
  {
    /*
      The following benchmarks were done with multiple files so the chance of
      results being cached was less likely. The purpose of keeping them here
      is for reference and documentation on why I chose to use preprocessor
      directives and native functions instead of streams or files.

      Benchmarks:
        C++ Code (Bad way):
          std::ios::sync_with_stdio(false);
          std::ifstream in(this->m_original_file, std::ios::binary);
          std::ofstream out(filename, std::ios::binary);

          std::copy(std::istreambuf_iterator<char>(in), std::istreambuf_iterator<char>(), std::ostreambuf_iterator<char>(out));

          out.close();
          in.close();

        Results:
          CLOCKS_PER_SEC: 1000
          TIME          : 237464
          TIME (SEC)    : 237.464 (3 min 57 sec)

        C++ Code (Better way):
          std::ios::sync_with_stdio(false);
          std::ifstream in(this->m_original_file, std::ios::binary);
          std::ofstream out(filename, std::ios::binary);

          out << in.rdbuf();

          out.close();
          in.close();

        Results:
          CLOCKS_PER_SEC: 1000
          TIME          : 24765
          TIME (SEC)    : 24.765 (24 sec 765 msec)

        Plain C Code (Dangerous way):
          FILE *in = fopen(this->m_original_file.c_str(), "rb");
          FILE *out = fopen(filename.c_str(), "wb");
          char *buf = new char[1024]();
          int read = 0;

          while ((read = fread(buf, 1, 1024, in) == 1024))
          {
            fwrite(buf, 1, 1024, out);
          }

          fwrite(buf, 1, read, out);

          fclose(out);
          fclose(in);

        Results:
          CLOCKS_PER_SEC: 1000
          TIME          : 2965
          TIME (SEC)    : 2.965 (3 sec)

        Dynamic Platform Specific Code (Windows)

          Results:
          CLOCKS_PER_SEC: 1000
          TIME          : 324
          TIME (SEC)    : 0.324 (324 msec)
    */
#ifdef _WIN32
    //  Windows
    CopyFileA(infile.c_str(), outfile.c_str(), false);
#elif __APPLE__
    //  OSX
    copyfile(infile.c_str(), outfile.c_str(), NULL, COPYFILE_DATA);
#elif __linux
    //  Linux
    stuct stat stat_buf;
    int in_fd, out_fd;
    offset_t offset = 0;

    in_fd = open(infile.c_str(), O_RDONLY);
    fstat(in_fd, &stat_buf);
    out_fd = open(outfile.c_str(), O_WRONLY | O_CREAT, stat_buf.st_mode);

    sendfile(out_fd, in_fd, &offset, stat_buf.st_size);

    close(out_fd);
    close(in_fd);
#endif
  }

  /*
    Determines whether a file exists. Fails if given path is not a file or doesn't exist.

    @return True if the file is both a file and exists.
  */
  inline bool exists(std::string filename)
  {
    struct stat buffer;
    return (stat(filename.c_str(), &buffer) == 0 && (buffer.st_mode & S_IFREG));
  }

  /*
	Gets the size of a file without streams.

	@return File size in bytes or 0 on failure.
  */
  inline uint64_t filesize(std::string path)
  {
	  struct stat buffer;
	  return stat(path.c_str(), &buffer) == 0 ? buffer.st_size : 0;
  }

  inline int mkdir(std::string path)
  {
#ifdef _WIN32
    //  Windows
    return _mkdir(path.c_str());
#else
    return mkdir(path.c_str());
#endif
  }

  inline int chdir(std::string path)
  {
#ifdef _WIN32
    //  Windows
    return _chdir(path.c_str());
#else
    return chdir(path.c_str());
#endif
  }
}


#endif
