#include "gba.h"

GBACorruption::GBACorruption()
{

}

GBACorruption::GBACorruption(std::string filename)
{
  initialize(filename);
}

GBACorruption::~GBACorruption()
{

}

void GBACorruption::initialize(std::string filename)
{
  this->read(filename);
  this->header = std::make_unique<GBAHeader>(ref(this->rom));
  this->info = std::make_unique<GBACorruptionInfo>();

  if (!valid())
  {
    throw InvalidRomException();
  }
}

void GBACorruption::corrupt()
{
  std::cout << "Rom size: " << std::hex << rom.size() << std::dec << std::endl;

  uint32_t corruptions = 0;
  std::uniform_int_distribution<int> dist(0x00, 0xFF);

  for (uint32_t i = GBAHeader::MaxSize + info->start;    // Start at header + start value
    i <= this->rom.size() - info->step && i < info->end; // While in a valid range
    i += info->step)                                     // Increase by step size
  {
    if (info->type == CorruptionType::Shift)
    {
      //  If it's okay to put the other byte in this position then change it
      if (i + info->value < rom.size() && valid_byte(this->rom[i + info->value], i))
      {
        corruptions++;
        this->rom[i] = this->rom[i + info->value];
      }
    }
    else if (info->type == CorruptionType::Swap)
    {

      if (i + info->value < rom.size() &&
        valid_byte(this->rom[i + info->value], i) &&  //  If it's okay to put the other byte in this position
        valid_byte(this->rom[i], i + info->value))    //  And it's okay to put this byte in the other position
      {
        corruptions++;
        uint8_t temp = this->rom[i + info->value];
        this->rom[i] = temp;
        this->rom[i + info->value] = temp;
      }
    }
    else if (info->type == CorruptionType::Add)
    {
      //  If the new byte value is valid then do the corruption
      if (valid_byte(this->rom[i] + info->value, i))
      {
        corruptions++;
        this->rom[i] += info->value;
      }
    }
    else if (info->type == CorruptionType::Set)
    {
      //  If the set value can be placed here then do it
      if (valid_byte(info->value, i))
      {
        corruptions++;
        this->rom[i] = info->value;
      }
    }
    else if (info->type == CorruptionType::Random)
    {
      bool corrupted = false;

      //  Try up to 100 times to corrupt
      for (uint32_t retry = 0; !corrupted && retry < 100; retry++)
      {
        uint8_t rand = dist(this->random);
        if (valid_byte(rand, i))
        {
          corruptions++;
          this->rom[i] = rand;
          corrupted = true;
        }
      }
    }
    else if (info->type == CorruptionType::RotateLeft)
    {
      uint8_t rotate = Util::rol<uint8_t>(this->rom[i], info->value);

      if (valid_byte(rotate, i))
      {
        this->rom[i] = rotate;
        corruptions++;
      }
    }
    else if (info->type == CorruptionType::RotateRight)
    {
      uint8_t rotate = Util::ror<uint8_t>(this->rom[i], info->value);

      if (valid_byte(rotate, i))
      {
        this->rom[i] = rotate;
        corruptions++;
      }
    }
    else if (info->type == CorruptionType::LogicalAnd)
    {
      if (valid_byte(this->rom[i] & info->value, i))
      {
        this->rom[i] &= info->value;
        corruptions++;
      }
    }
    else if (info->type == CorruptionType::LogicalOr)
    {
      if (valid_byte(this->rom[i] | info->value, i))
      {
        this->rom[i] |= info->value;
        corruptions++;
      }
    }
    else if (info->type == CorruptionType::LogicalXor)
    {
      if (valid_byte(this->rom[i] ^ info->value, i))
      {
        this->rom[i] ^= info->value;
        corruptions++;
      }
    }
    else if (info->type == CorruptionType::LogicalComplement)
    {
      if (valid_byte(~this->rom[i], i))
      {
        this->rom[i] = ~this->rom[i];
        corruptions++;
      }
    }
    else
    {
      break;  //  No corruption selected, might as well quit.
    }
  }
  std::cout << "Replaced a total of " << corruptions << " bytes." << std::endl;
}

bool GBACorruption::valid()
{
  return header->valid();
}

bool GBACorruption::valid_byte(uint8_t byte, uint32_t location)
{
  //  Entire instruction is at location of the nearest 4 byte boundary
  uint32_t instruction = Util::read<uint32_t>(ref(rom), location - (location % 4));

  /*
  Format for ARM7 instructions:

  +-----------------------------------------------------+
  |31  28|27      20|19               8|7        4|3   0|
  |-----------------------------------------------------|
  | COND | OPCODE 1 | PARAMETERS/OTHER | OPCODE 2 | RM  |
  +-----------------------------------------------------+
  */

  uint8_t condition = (instruction & 0xF0000000) >> 24;
  uint8_t opcode1 = (instruction & 0x0FF00000) >> 20;
  uint8_t opcode2 = (instruction & 0x000000F0) >> 4;

  //  When opcode1 is between 0x80 and 0xE0 it will be jumps or load/store
  if (opcode1 >= 0x80 && opcode1 < 0xE0)
  {
    return false;
  }

  //  Every even opcode1 between 0x60 and 0x80 is a load/store
  if (opcode1 >= 0x60 && opcode1 < 0x80 && opcode2 % 2 == 0)
  {
    return false;
  }

  //  Opcode1 between 0x40 and 0x60 are all load/store
  if (opcode1 >= 0x40 && opcode1 < 0x60)
  {
    return false;
  }

  //  If opcode1 is between 0x00 and 0x20
  if (opcode1 >= 0x00 && opcode1 < 0x20)
  {
    if ((opcode2 & 0x0B) == 0x0B)
    {
      //  All opcode2 that end with 0x0B are load/store
      return false;
    }
    else if (opcode1 % 2 == 1 && ((opcode2 & 0x0D) == 0x0D || (opcode2 & 0x0F) == 0x0F))
    {
      //  All opcode2 that end with 0x0D or 0x0F and have an odd opcode1 are load/store
      return false;
    }
  }

  return true;
}

void GBACorruption::print_header()
{

}

void GBACorruption::parse(std::vector<std::string> args)
{
  info = std::make_unique<GBACorruptionInfo>();

  for (uint32_t i = 0; i < args.size(); i++)
  {
    auto& arg = args[i];

    if (arg == "-s" || arg == "--step")
    {
      info->step = Util::to_int32(args[++i]);
    }
    else if (arg == "-h" || arg == "--shift")
    {
      info->value = info->value = Util::to_int32(args[++i]);
      info->type = CorruptionType::Shift;
    }
    else if (arg == "-w" || arg == "--swap")
    {
      info->value = info->value = Util::to_int32(args[++i]);
      info->type = CorruptionType::Swap;
    }
    else if (arg == "-a" || arg == "--add")
    {
      info->value = Util::to_int32(args[++i]);
      info->type = CorruptionType::Add;
    }
    else if (arg == "-t" || arg == "--set")
    {
      info->value = Util::to_int32(args[++i]);
      info->type = CorruptionType::Set;
    }
    else if (arg == "-r" || arg == "--random")
    {
      info->type = CorruptionType::Random;
    }
    else if (arg == "--rotate-left")
    {
      info->value = Util::to_int32(args[++i]);
      info->type = CorruptionType::RotateLeft;
    }
    else if (arg == "--rotate-right")
    {
      info->value = Util::to_int32(args[++i]);
      info->type = CorruptionType::RotateRight;
    }
    else if (arg == "--logical-and")
    {
      info->value = Util::to_int32(args[++i]);
      info->type = CorruptionType::LogicalAnd;
    }
    else if (arg == "--logical-or")
    {
      info->value = Util::to_int32(args[++i]);
      info->type = CorruptionType::LogicalOr;
    }
    else if (arg == "--logical-xor")
    {
      info->value = Util::to_int32(args[++i]);
      info->type = CorruptionType::LogicalXor;
    }
    else if (arg == "--logical-complement")
    {
      info->type = CorruptionType::LogicalComplement;
    }
    else if (arg == "-b" || arg == "--start")
    {
      info->start = Util::to_int32(args[++i]);
    }
    else if (arg == "-e" || arg == "--stop")
    {
      info->end = Util::to_int32(args[++i]);
    }
    else if (arg == "-o" || arg == "--out")
    {
      info->outfile = args[i + 1];
      i++;
    }
    else
    {
      throw InvalidArgumentException(arg + " is not a recognized flag.");
    }
  }
}

void GBACorruption::save(std::string filename)
{
  std::string format = ".gba";

  //  If the filename doesn't include the format extension then add it.
  if (filename.length() <= 4 || filename.compare(filename.length() - format.length(), format.length(), format) != 0)
  {
    filename += format;
  }

  try
  {
    if (info->outfile != "")
    {
      this->rom_file.open(info->outfile, std::ios::out | std::ios::binary);
    }
    else
    {
      this->rom_file.open(filename, std::ios::out | std::ios::binary);
    }

    rom_file.write(reinterpret_cast<char *>(&rom[0]), rom.size());
    //std::copy(rom.begin(), rom.end(), std::ostream_iterator<uint8_t>(this->rom_file));
    this->rom_file.close();

    this->save_name = filename;
  }
  catch (InvalidFileNameException e)
  {
    std::cout << e.what() << std::endl;
    exit(EXIT_FAILURE);
  }
}
