#include "gbc.h"

GBCCorruption::GBCCorruption() { }

GBCCorruption::GBCCorruption(std::string filename)
{
  this->initialize(filename);
}

GBCCorruption::~GBCCorruption() { }

void GBCCorruption::initialize(std::string filename)
{
  // Read the rom file
  this->read(filename);

  this->header = std::make_unique<GBCHeader>(ref(this->rom));

  // If the rom is not valid then throw an exception
  if (!this->valid())
  {
    throw InvalidGBCRomException();
  }

  // Rom is valid, set up variables.
}

void GBCCorruption::corrupt()
{
  std::cout << "Rom size: " << std::hex << rom.size() << std::dec << std::endl;

  uint32_t corruptions = 0;
  std::uniform_int_distribution<int> dist(0x00, 0xFF);

  for (uint32_t i = GBCHeader::Start + info->start;         // Start at header + start value
      i <= this->rom.size() - info->step && i < info->end;  // While in a valid range
      i += info->step)                                      // Increase by step size
  {
    if (info->type == CorruptionType::Shift)
    {
      //  If it's okay to put the other byte in this position then change it
      if (i + info->value && valid_byte(this->rom[i + info->value], i))
      {
        corruptions++;
        this->rom[i] = this->rom[i + info->value];
      }
    }
    else if (info->type == CorruptionType::Swap)
    {

      if (i + info->value &&
          valid_byte(this->rom[i + info->value], i) &&  //  If it's okay to put the other byte in this position
          valid_byte(this->rom[i], i + info->value))    //  And it's okay to put this byte in the other position
      {
        corruptions++;
        uint8_t temp = this->rom[i + info->value];
        this->rom[i] = temp;
        this->rom[i + info->value] = temp;
      }
    }
    else if (info->type == CorruptionType::Add)
    {
      //  If the new byte value is valid then do the corruption
      if (valid_byte(this->rom[i] + info->value, i))
      {
        corruptions++;
        this->rom[i] += info->value;
      }
    }
    else if (info->type == CorruptionType::Set)
    {
      //  If the set value can be placed here then do it
      if (valid_byte(info->value, i))
      {
        corruptions++;
        this->rom[i] = info->value;
      }
    }
    else if (info->type == CorruptionType::Random)
    {
      bool corrupted = false;

      //  Try up to 100 times to corrupt
      for (uint32_t retry = 0; !corrupted && retry < 100; retry++)
      {
        uint8_t rand = dist(this->random);
        if (valid_byte(rand, i))
        {
          corruptions++;
          this->rom[i] = rand;
          corrupted = true;
        }
      }
    }
    else if (info->type == CorruptionType::RotateLeft)
    {
      uint8_t rotate = Util::rol<uint8_t>(this->rom[i], info->value);

      if (valid_byte(rotate, i))
      {
        this->rom[i] = rotate;
        corruptions++;
      }
    }
    else if (info->type == CorruptionType::RotateRight)
    {
      uint8_t rotate = Util::ror<uint8_t>(this->rom[i], info->value);

      if (valid_byte(rotate, i))
      {
        this->rom[i] = rotate;
        corruptions++;
      }
    }
    else if (info->type == CorruptionType::LogicalAnd)
    {
      if (valid_byte(this->rom[i] & info->value, i))
      {
        this->rom[i] &= info->value;
        corruptions++;
      }
    }
    else if (info->type == CorruptionType::LogicalOr)
    {
      if (valid_byte(this->rom[i] | info->value, i))
      {
        this->rom[i] |= info->value;
        corruptions++;
      }
    }
    else if (info->type == CorruptionType::LogicalXor)
    {
      if (valid_byte(this->rom[i] ^ info->value, i))
      {
        this->rom[i] ^= info->value;
        corruptions++;
      }
    }
    else if (info->type == CorruptionType::LogicalComplement)
    {
      if (valid_byte(~this->rom[i], i))
      {
        this->rom[i] = ~this->rom[i];
        corruptions++;
      }
    }
    else
    {
      break;  //  No corruption selected, might as well quit.
    }
  }

  std::cout << "Replaced a total of " << corruptions << " bytes." << std::endl;
}

void GBCCorruption::print_header()
{
  std::cout << std::hex;
  std::cout << "Entry point:  " << static_cast<int>(this->header->entry()) << std::endl;
  std::cout << "Valid:        " << static_cast<bool>(this->header->valid_logo()) << std::endl;
  std::cout << "Name:         " << this->header->name() << std::endl;
  std::cout << "Manufacturer: " << this->header->manufacturer() << std::endl;
  std::cout << "CGB Enabled:  " << static_cast<int>(this->header->cgb()) << std::endl;
  std::cout << "Company:      " << this->header->company() << std::endl;
  std::cout << "SGB Enabled:  " << static_cast<int>(this->header->sgb()) << std::endl;
  std::cout << "Cart Type:    " << static_cast<int>(this->header->type()) << std::endl;
  std::cout << "Rom Size:     " << static_cast<int>(this->header->rom_size()) << std::endl;
  std::cout << "Ram Size:     " << static_cast<int>(this->header->ram_size()) << std::endl;
  std::cout << "Japanese:     " << static_cast<int>(this->header->destination()) << std::endl;
  std::cout << "Old License:  " << static_cast<int>(this->header->old_license()) << std::endl;
  std::cout << std::dec;
}

bool GBCCorruption::valid()
{
  return this->header->valid_logo();
}

bool GBCCorruption::valid_byte(uint8_t byte, uint32_t location)
{
  //  Map of branch opcodes and their size
  //  std::map<opcode, size>
  static std::map<uint8_t, uint8_t> invalid_branch{
    { 0xC2, 3 }, { 0xD2, 3 }, { 0xC3, 3 }, { 0xC4, 3 },
    { 0xD4, 3 }, { 0xCA, 3 }, { 0xDA, 3 }, { 0XCC, 3 },
    { 0xCD, 3 }, { 0xDC, 3 }, { 0x20, 2 }, { 0x30, 2 },
    { 0x18, 2 }, { 0x28, 2 }, { 0x38, 2 }, { 0xE9, 1 }
  };

  //  List of stack operations (push/pop)
  static std::vector<uint8_t> invalid_stack{
    0xC1, 0xD1, 0xE1, 0xF1, 0xC5, 0xD5, 0xE5, 0xF5
  };

  //  List of 1 byte return opcodes
  static std::vector<uint8_t> invalid_ret{
    0xC0, 0xD0, 0xC7, 0xD7, 0xE7, 0xF7, 0xC8,
    0xD8, 0xC9, 0xD9, 0xCF, 0xDF, 0xEF, 0xFF
  };

  if (std::find(invalid_ret.begin(), invalid_ret.end(), this->rom[location]) != invalid_ret.end() ||
      std::find(invalid_ret.begin(), invalid_ret.end(), byte) != invalid_ret.end())
  {
    //  If replaced byte or current byte is a 1 byte ret opcode then it isn't safe to replace
    return false;
  }

  //  Check if we are a branch opcode or the parameter of a branch opcode
  for (auto& branch : invalid_branch)
  {
    auto opcode = branch.first;
    auto size = branch.second;

    //  Check up to [size] bytes behind us to make sure this location or byte value isn't a branch parameter
    for (int i = 0; i < size; i++)
    {
      //  If it is then it is not valid
      if (this->rom[location - i] == opcode || byte == opcode)
      {
        return false;
      }
    }
  }

  return true;
}

void GBCCorruption::parse(std::vector<std::string> args)
{
  info = std::make_unique<GBCCorruptionInfo>();

  for (uint32_t i = 0; i < args.size(); i++)
  {
    auto& arg = args[i];

    if (arg == "-s" || arg == "--step")
    {
      info->step = Util::to_int32(args[++i]);
    }
    else if (arg == "-h" || arg == "--shift")
    {
      info->value = info->value = Util::to_int32(args[++i]);
      info->type = CorruptionType::Shift;
    }
    else if (arg == "-w" || arg == "--swap")
    {
      info->value = info->value = Util::to_int32(args[++i]);
      info->type = CorruptionType::Swap;
    }
    else if (arg == "-a" || arg == "--add")
    {
      info->value = Util::to_int32(args[++i]);
      info->type = CorruptionType::Add;
    }
    else if (arg == "-t" || arg == "--set")
    {
      info->value = Util::to_int32(args[++i]);
      info->type = CorruptionType::Set;
    }
    else if (arg == "-r" || arg == "--random")
    {
      info->type = CorruptionType::Random;
    }
    else if (arg == "--rotate-left")
    {
      info->value = Util::to_int32(args[++i]);
      info->type = CorruptionType::RotateLeft;
    }
    else if (arg == "--rotate-right")
    {
      info->value = Util::to_int32(args[++i]);
      info->type = CorruptionType::RotateRight;
    }
    else if (arg == "--logical-and")
    {
      info->value = Util::to_int32(args[++i]);
      info->type = CorruptionType::LogicalAnd;
    }
    else if (arg == "--logical-or")
    {
      info->value = Util::to_int32(args[++i]);
      info->type = CorruptionType::LogicalOr;
    }
    else if (arg == "--logical-xor")
    {
      info->value = Util::to_int32(args[++i]);
      info->type = CorruptionType::LogicalXor;
    }
    else if (arg == "--logical-complement")
    {
      info->type = CorruptionType::LogicalComplement;
    }
    else if (arg == "-b" || arg == "--start")
    {
      info->start = Util::to_int32(args[++i]);
    }
    else if (arg == "-e" || arg == "--stop")
    {
      info->end = Util::to_int32(args[++i]);
    }
    else if (arg == "-o" || arg == "--out")
    {
      info->outfile = args[i + 1];
      i++;
    }
    else
    {
      throw InvalidArgumentException(arg + " is not a recognized flag.");
    }
  }
}

void GBCCorruption::save(std::string filename)
{
  std::string format = ".gbc";

  //  If the filename doesn't include the format extension then add it.
  if (filename.length() <= 4 || filename.compare(filename.length() - format.length(), format.length(), format) != 0)
  {
    filename += format;
  }

  try
  {
    if (info->outfile != "")
    {
      this->rom_file.open(info->outfile, std::ios::out | std::ios::binary);
    }
    else
    {
      this->rom_file.open(filename, std::ios::out | std::ios::binary);
    }

    std::copy(rom.begin(), rom.end(), std::ostream_iterator<uint8_t>(this->rom_file));
    this->rom_file.close();

    this->save_name = filename;
  }
  catch (InvalidFileNameException e)
  {
    std::cout << e.what() << std::endl;
    exit(EXIT_FAILURE);
  }
}