#include "genesis.h"

GenesisCorruption::GenesisCorruption()
{

}

GenesisCorruption::GenesisCorruption(std::string filename)
{
  this->initialize(filename);
}

GenesisCorruption::~GenesisCorruption()
{

}

void GenesisCorruption::initialize(std::string filename)
{
  // Read the rom file
  this->read(filename);
  this->info = std::make_unique<GenesisCorruptionInfo>();
  this->header = std::make_unique<GenesisHeader>(ref(this->rom));

  // If the rom is not valid then throw an exception
  if (!this->valid())
  {
    throw InvalidGenesisRomException();
  }

  // Rom is valid, set up variables.
}

void GenesisCorruption::corrupt()
{
  debug::cout << "Rom size: " << std::hex << rom.size() << std::dec << std::endl;

  uint32_t corruptions = 0;
  std::uniform_int_distribution<int> dist(0x00, 0xFF);

  for (uint32_t i = this->header->begin() + info->start;    // Start at header + start value
    i <= this->rom.size() - info->step && i < info->end; // While in a valid range
    i += info->step)                                     // Increase by step size
  {
    if (info->type == CorruptionType::Shift)
    {
      //  If it's okay to put the other byte in this position then change it
      if (i + info->value < rom.size() && valid_byte(this->rom[i + info->value], i))
      {
        corruptions++;
        this->rom[i] = this->rom[i + info->value];
      }
    }
    else if (info->type == CorruptionType::Swap)
    {

      if (i + info->value < rom.size() &&
        valid_byte(this->rom[i + info->value], i) &&  //  If it's okay to put the other byte in this position
        valid_byte(this->rom[i], i + info->value))    //  And it's okay to put this byte in the other position
      {
        corruptions++;
        uint8_t temp = this->rom[i + info->value];
        this->rom[i] = temp;
        this->rom[i + info->value] = temp;
      }
    }
    else if (info->type == CorruptionType::Add)
    {
      //  If the new byte value is valid then do the corruption
      if (valid_byte(this->rom[i] + info->value, i))
      {
        corruptions++;
        this->rom[i] += info->value;
      }
    }
    else if (info->type == CorruptionType::Set)
    {
      //  If the set value can be placed here then do it
      if (valid_byte(info->value, i))
      {
        corruptions++;
        this->rom[i] = info->value;
      }
    }
    else if (info->type == CorruptionType::Random)
    {
      bool corrupted = false;

      //  Try up to 100 times to corrupt
      for (uint32_t retry = 0; !corrupted && retry < 100; retry++)
      {
        uint8_t rand = dist(this->random);
        if (valid_byte(rand, i))
        {
          corruptions++;
          this->rom[i] = rand;
          corrupted = true;
        }
      }
    }
    else if (info->type == CorruptionType::RotateLeft)
    {
      uint8_t rotate = Util::rol<uint8_t>(this->rom[i], info->value);

      if (valid_byte(rotate, i))
      {
        this->rom[i] = rotate;
        corruptions++;
      }
    }
    else if (info->type == CorruptionType::RotateRight)
    {
      uint8_t rotate = Util::ror<uint8_t>(this->rom[i], info->value);

      if (valid_byte(rotate, i))
      {
        this->rom[i] = rotate;
        corruptions++;
      }
    }
    else if (info->type == CorruptionType::LogicalAnd)
    {
      if (valid_byte(this->rom[i] & info->value, i))
      {
        this->rom[i] &= info->value;
        corruptions++;
      }
    }
    else if (info->type == CorruptionType::LogicalOr)
    {
      if (valid_byte(this->rom[i] | info->value, i))
      {
        this->rom[i] |= info->value;
        corruptions++;
      }
    }
    else if (info->type == CorruptionType::LogicalXor)
    {
      if (valid_byte(this->rom[i] ^ info->value, i))
      {
        this->rom[i] ^= info->value;
        corruptions++;
      }
    }
    else if (info->type == CorruptionType::LogicalComplement)
    {
      if (valid_byte(~this->rom[i], i))
      {
        this->rom[i] = ~this->rom[i];
        corruptions++;
      }
    }
    else
    {
      break;  //  No corruption selected, might as well quit.
    }
  }
  debug::cout << "Replaced a total of " << corruptions << " bytes." << std::endl;
}

/*
Determines whether or not a given byte is valid to include into the rom.

@param byte - The byte that needs to be checked
@param location - The offset that the byte will be placed at in the rom

@return true if the byte can be added at @location in the rom.
*/
bool GenesisCorruption::valid_byte(uint8_t byte, uint32_t location)
{
  if (is_branch(Util::read_big<uint16_t>(rom, location)) ||
    is_branch(Util::read_big<uint16_t>(rom, location - 1)))
  {
    return true;
  }

  return true;
}

bool GenesisCorruption::is_branch(uint16_t instruction)
{
  uint8_t hi_byte = (instruction & 0xFF00) >> 8;
  uint8_t hi_nybble = (instruction & 0xF000) >> 12;

  return (hi_nybble == 0x06 ||  //  Bcc
          hi_byte == 0x61 || // BSR
          hi_byte == 0x60 || // BRA
          hi_byte == 0x4E75 ||  //  RTS
          hi_byte == 0x4E77 ||  //  RTR
          hi_byte == 0x4E74 ||  //  RTD
          (instruction & 0xFFC0) == 0x4E80 || //  JSR
          (instruction & 0xFFC0) == 0x4EC0 || //  JMP
          (instruction & 0xF0F8) == 0x50C8);  //DBcc
}

/*
Prints out basic header information for debugging purposes.
*/
void GenesisCorruption::print_header()
{
}

/*
Determines whether or not the currently loaded rom
is a valid Genesis rom by checking if the header exists.

@return true for a valid header or false otherwise.
*/
bool GenesisCorruption::valid()
{
  return this->header->valid();
}

void GenesisCorruption::parse(std::vector<std::string> args)
{
  info = std::make_unique<GenesisCorruptionInfo>();

  for (uint32_t i = 0; i < args.size(); i++)
  {
    auto& arg = args[i];

    if (arg == "-s" || arg == "--step")
    {
      info->step = Util::to_int32(args[++i]);
    }
    else if (arg == "-h" || arg == "--shift")
    {
      info->value = info->value = Util::to_int32(args[++i]);
      info->type = CorruptionType::Shift;
    }
    else if (arg == "-w" || arg == "--swap")
    {
      info->value = info->value = Util::to_int32(args[++i]);
      info->type = CorruptionType::Swap;
    }
    else if (arg == "-a" || arg == "--add")
    {
      info->value = Util::to_int32(args[++i]);
      info->type = CorruptionType::Add;
    }
    else if (arg == "-t" || arg == "--set")
    {
      info->value = Util::to_int32(args[++i]);
      info->type = CorruptionType::Set;
    }
    else if (arg == "-r" || arg == "--random")
    {
      info->type = CorruptionType::Random;
    }
    else if (arg == "--rotate-left")
    {
      info->value = Util::to_int32(args[++i]);
      info->type = CorruptionType::RotateLeft;
    }
    else if (arg == "--rotate-right")
    {
      info->value = Util::to_int32(args[++i]);
      info->type = CorruptionType::RotateRight;
    }
    else if (arg == "--logical-and")
    {
      info->value = Util::to_int32(args[++i]);
      info->type = CorruptionType::LogicalAnd;
    }
    else if (arg == "--logical-or")
    {
      info->value = Util::to_int32(args[++i]);
      info->type = CorruptionType::LogicalOr;
    }
    else if (arg == "--logical-xor")
    {
      info->value = Util::to_int32(args[++i]);
      info->type = CorruptionType::LogicalXor;
    }
    else if (arg == "--logical-complement")
    {
      info->type = CorruptionType::LogicalComplement;
    }
    else if (arg == "-b" || arg == "--start")
    {
      info->start = Util::to_int32(args[++i]);
    }
    else if (arg == "-e" || arg == "--stop")
    {
      info->end = Util::to_int32(args[++i]);
    }
    else if (arg == "-o" || arg == "--out")
    {
      info->outfile = args[i + 1];
      i++;
    }
    else
    {
      throw InvalidArgumentException(arg + " is not a recognized flag.");
    }
  }
}

/*
Saves the current ROM to a file, automatically taking into account
the extension.

@param filename - The file to save to. Does not require extension.
*/
void GenesisCorruption::save(std::string filename)
{
  std::string format = ".md";

  //  If the filename doesn't include the format extension then add it.
  if (filename.length() <= 4 || filename.compare(filename.length() - format.length(), format.length(), format) != 0)
  {
    filename += format;
  }

  //  Fix the checksum
  uint16_t cs = checksum();
  rom[0x18E] = (cs & 0xFF00) >> 8;
  rom[0x18F] = (cs & 0xFF);

  try
  {
    if (info->outfile != "")
    {
      this->rom_file.open(info->outfile, std::ios::out | std::ios::binary);
    }
    else
    {
      this->rom_file.open(filename, std::ios::out | std::ios::binary);
    }

    std::copy(rom.begin(), rom.end(), std::ostream_iterator<uint8_t>(this->rom_file));
    this->rom_file.close();

    this->save_name = filename;
  }
  catch (InvalidFileNameException e)
  {
    std::cout << e.what() << std::endl;
    exit(EXIT_FAILURE);
  }
}

uint16_t GenesisCorruption::checksum()
{
  uint16_t sum = 0;

  for (uint32_t i = 0x200; i < rom.size();)
  {
    sum += rom[i++] * 0x100;
    sum += rom[i++];
  }

  debug::cout << "Checksum: " << std::hex << sum << std::dec << std::endl;

  return sum;
}