#include "n64.h"

N64Corruption::N64Corruption()
{

}

N64Corruption::N64Corruption(std::string filename)
{
  this->initialize(filename);
}

N64Corruption::~N64Corruption()
{

}

void N64Corruption::initialize(std::string filename)
{
  // Read the rom file
  this->read(filename);

  this->header = std::make_unique<N64Header>(ref(this->rom));

  // If the rom is not valid then throw an exception
  if (!this->valid())
  {
    throw InvalidN64RomException();
  }

  // Rom is valid, set up variables.
}

void N64Corruption::corrupt()
{
  uint64_t corruptions = 0;

  std::uniform_int_distribution<int> dist(0x00, 0xFF);

  for (uint32_t i = N64Corruption::Start + info->start;     // Start at header + start value
       i <= this->rom.size() - info->step && i < info->end; // While in a valid range
       i += info->step)                                     // Increase by step size
  {
    if (info->type == CorruptionType::Shift)
    {
      //  If it's okay to put the other byte in this position then change it
      if (i + info->value < rom.size() && valid_byte(this->rom[i + info->value], i))
      {
        corruptions++;
        this->rom[i] = this->rom[i + info->value];
      }
    }
    else if (info->type == CorruptionType::Swap)
    {

      if (i + info->value < rom.size() &&
        valid_byte(this->rom[i + info->value], i) &&  //  If it's okay to put the other byte in this position
        valid_byte(this->rom[i], i + info->value))    //  And it's okay to put this byte in the other position
      {
        corruptions++;
        uint8_t temp = this->rom[i + info->value];
        this->rom[i] = temp;
        this->rom[i + info->value] = temp;
      }
    }
    else if (info->type == CorruptionType::Add)
    {
      //  If the new byte value is valid then do the corruption
      if (valid_byte(this->rom[i] + info->value, i))
      {
        corruptions++;
        this->rom[i] += info->value;
      }
    }
    else if (info->type == CorruptionType::Set)
    {
      //  If the set value can be placed here then do it
      if (valid_byte(info->value, i))
      {
        corruptions++;
        this->rom[i] = info->value;
      }
    }
    else if (info->type == CorruptionType::Random)
    {
      bool corrupted = false;

      //  Try up to 100 times to corrupt
      for (uint32_t retry = 0; !corrupted && retry < 100; retry++)
      {
        uint8_t rand = dist(this->random);
        if (valid_byte(rand, i))
        {
          corruptions++;
          this->rom[i] = rand;
          corrupted = true;
        }
      }
    }
    else if (info->type == CorruptionType::RotateLeft)
    {
      uint8_t rotate = Util::rol<uint8_t>(this->rom[i], info->value);

      if (valid_byte(rotate, i))
      {
        this->rom[i] = rotate;
        corruptions++;
      }
    }
    else if (info->type == CorruptionType::RotateRight)
    {
      uint8_t rotate = Util::ror<uint8_t>(this->rom[i], info->value);

      if (valid_byte(rotate, i))
      {
        this->rom[i] = rotate;
        corruptions++;
      }
    }
    else if (info->type == CorruptionType::LogicalAnd)
    {
      if (valid_byte(this->rom[i] & info->value, i))
      {
        this->rom[i] &= info->value;
        corruptions++;
      }
    }
    else if (info->type == CorruptionType::LogicalOr)
    {
      if (valid_byte(this->rom[i] | info->value, i))
      {
        this->rom[i] |= info->value;
        corruptions++;
      }
    }
    else if (info->type == CorruptionType::LogicalXor)
    {
      if (valid_byte(this->rom[i] ^ info->value, i))
      {
        this->rom[i] ^= info->value;
        corruptions++;
      }
    }
    else if (info->type == CorruptionType::LogicalComplement)
    {
      if (valid_byte(~this->rom[i], i))
      {
        this->rom[i] = ~this->rom[i];
        corruptions++;
      }
    }
    else
    {
      break;  //  No corruption selected, might as well quit.
    }
  }
  std::cout << "Replaced a total of " << corruptions << " bytes." << std::endl;
}

void N64Corruption::print_header()
{
  std::cout << std::hex;
  std::cout << "Domain:          " << this->header->domain() << std::endl;
  std::cout << "Clock Rate:      " << this->header->clock_rate() << std::endl;
  std::cout << "Program Counter: " << this->header->program_counter() << std::endl;
  std::cout << "Release:         " << this->header->release() << std::endl;
  std::cout << "CRC1:            " << this->header->crc1() << std::endl;
  std::cout << "CRC2:            " << this->header->crc2() << std::endl;
  std::cout << "Name:            " << this->header->name() << std::endl;
  std::cout << "Format:          " << this->header->format() << std::endl;
  std::cout << "ID:              " << this->header->id() << std::endl;
  std::cout << "Region:          " << this->header->region() << std::endl;
  std::cout << "Version:         " << this->header->version() << std::endl;
  std::cout << std::dec;
}

bool N64Corruption::valid()
{
  return true;
}

bool N64Corruption::valid_byte(uint8_t byte, uint32_t location)
{
  //  Get instruction which is bound to a 4 byte alignment
  uint32_t instruction = Util::read<uint32_t>(ref(rom), location - (location % 4));

  //  Get first opcode which is between bits 31 and 26 ( 6 bits total )
  uint8_t opcode = (instruction >> 24) & 0xFC;
  uint8_t high = opcode & 0x38;
  uint8_t low = opcode & 0x07;

  /*
        31---------26---------------------------------------------------0
        |  opcode   |                                                   |
        ------6----------------------------------------------------------
        |--000--|--001--|--010--|--011--|--100--|--101--|--110--|--111--| lo
    000 | *1    | *2    | J     | JAL   | BEQ   | BNE   | BLEZ  | BGTZ  |
    001 | ADDI  | ADDIU | SLTI  | SLTIU | ANDI  | ORI   | XORI  | LUI   |
    010 | *3    | *4    | *5    |       | BEQL  | BNEL  | BLEZL | BGTZL |
    011 | DADDI |DADDIU |  LDL  |  LDR  |       |       |       |       |
    100 | LB    | LH    | LWL   | LW    | LBU   | LHU   | LWR   | LWU   |
    101 | SB    | SH    | SWL   | SW    | SDL   | SDR   | SWR   | CACHE |
    110 | LL    | LWC1  | LWC2  |       | LLD   | LDC1  | LDC2  | LD    |
    111 | SC    | SWC1  | SWC2  |       | SCD   | SDC1  | SDC2  | SD    |
     hi |-------|-------|-------|-------|-------|-------|-------|-------|
         *1 = SPECIAL, see SPECIAL list      *2 = REGIMM, see REGIMM list
         *3 = COP0                  *4 = COP1                   *5 = COP2
  */

  if (opcode == 0x00)       // *1 SPECIAL
  {
    uint8_t function = instruction & 0x3F;
    //  Block the following opcodes:
    //  SLL | SRL | SRA | SLLV | SRLV | SRAV | JR | JALR | SYSCALL | BREAK | SYNC
    if (function <= 0x0F)
    {
      return false;
    }
  }
  else if (opcode == 0x01)  // *2 REGIMM
  {
    return false; // No acceptable instructions under REGIMM
  }
  else if (opcode == 0x10)  // *3 COP0
  {
    //  Format is 5 bits from 25 to 21 in the instruction
    uint8_t format = instruction & 0x3E00000;
    uint8_t fmthigh = format & 0x18;
    uint8_t fmtlow = format & 0x07;

    if (fmthigh == 1)       //  *1 BC Instruction
    {
      return false; // No valid BC instructions
    }
    else if (fmthigh == 2)  //  *2 TLB Instruction
    {
      return false; // No valid TLB instructions
    }
  }
  else if (opcode == 0x11)  // *4 COP1
  {
    //  Format is 5 bits from 25 to 21 in the instruction
    uint8_t format = instruction & 0x3E00000;

    if (format == 8)
    {
      return false; // No valid BC instructions
    }
  }
  else if (opcode == 0x12)  // *4 COP2
  {
    //  No documentation on COP2 instructions from the resource I was given.
    return false; // In the mean while, return false to be safe.
  }

  return true;
}

void N64Corruption::parse(std::vector<std::string> args)
{
  info = std::make_unique<N64CorruptionInfo>();

  for (uint32_t i = 0; i < args.size(); i++)
  {
    auto& arg = args[i];

    if (arg == "-s" || arg == "--step")
    {
      info->step = Util::to_int32(args[++i]);
    }
    else if (arg == "-h" || arg == "--shift")
    {
      info->value = info->value = Util::to_int32(args[++i]);
      info->type = CorruptionType::Shift;
    }
    else if (arg == "-w" || arg == "--swap")
    {
      info->value = info->value = Util::to_int32(args[++i]);
      info->type = CorruptionType::Swap;
    }
    else if (arg == "-a" || arg == "--add")
    {
      info->value = Util::to_int32(args[++i]);
      info->type = CorruptionType::Add;
    }
    else if (arg == "-t" || arg == "--set")
    {
      info->value = Util::to_int32(args[++i]);
      info->type = CorruptionType::Set;
    }
    else if (arg == "-r" || arg == "--random")
    {
      info->type = CorruptionType::Random;
    }
    else if (arg == "--rotate-left")
    {
      info->value = Util::to_int32(args[++i]);
      info->type = CorruptionType::RotateLeft;
    }
    else if (arg == "--rotate-right")
    {
      info->value = Util::to_int32(args[++i]);
      info->type = CorruptionType::RotateRight;
    }
    else if (arg == "--logical-and")
    {
      info->value = Util::to_int32(args[++i]);
      info->type = CorruptionType::LogicalAnd;
    }
    else if (arg == "--logical-or")
    {
      info->value = Util::to_int32(args[++i]);
      info->type = CorruptionType::LogicalOr;
    }
    else if (arg == "--logical-xor")
    {
      info->value = Util::to_int32(args[++i]);
      info->type = CorruptionType::LogicalXor;
    }
    else if (arg == "--logical-complement")
    {
      info->type = CorruptionType::LogicalComplement;
    }
    else if (arg == "-b" || arg == "--start")
    {
      info->start = Util::to_int32(args[++i]);
    }
    else if (arg == "-e" || arg == "--stop")
    {
      info->end = Util::to_int32(args[++i]);
    }
    else if (arg == "-o" || arg == "--out")
    {
      info->outfile = args[i + 1];
      i++;
    }
    else
    {
      throw InvalidArgumentException(arg + " is not a recognized flag.");
    }
  }
}

void N64Corruption::save(std::string filename)
{
  std::string format = ".z64";

  //  If the filename doesn't include the format extension then add it.
  if (filename.length() <= 4 || filename.compare(filename.length() - format.length(), format.length(), format) != 0)
  {
    filename += format;
  }

  try
  {
    if (info->outfile != "")
    {
      this->rom_file.open(info->outfile, std::ios::out | std::ios::binary);
    }
    else
    {
      this->rom_file.open(filename, std::ios::out | std::ios::binary);
    }

    std::copy(rom.begin(), rom.end(), std::ostream_iterator<uint8_t>(this->rom_file));
    this->rom_file.close();

    this->save_name = filename;
  }
  catch (InvalidFileNameException e)
  {
    std::cout << e.what() << std::endl;
    exit(EXIT_FAILURE);
  }
}

void N64Corruption::convert_z64(std::string filename)
{
  uint32_t size = static_cast<uint32_t>(boost::filesystem::file_size(filename));
  std::string extension = boost::filesystem::extension(filename);
  std::vector<uint8_t> data(size);

  std::ifstream ifs(filename, std::ios::binary);
  std::ofstream ofs(filename.substr(0, filename.length() - 3) + "z64", std::ios::binary);

  ifs.read(reinterpret_cast<char *>(&data[0]), size);

  if (extension == ".n64")
  {
    for (uint32_t i = 0; i < size; i += 2)
    {
      uint8_t temp = data[i];
      data[i] = data[i + 1];
      data[i + 1] = temp;
    }
  }
  else if (extension == ".v64")
  {

  }

  ofs.write(reinterpret_cast<char *>(&data[0]), size);

  ofs.close();
  ifs.close();
}