#include "snes.h"

SNESCorruption::SNESCorruption()
{

}

SNESCorruption::SNESCorruption(std::string filename)
{
  this->initialize(filename);
}

SNESCorruption::~SNESCorruption()
{

}

void SNESCorruption::initialize(std::string filename)
{
  // Read the rom file
  this->read(filename);
  this->info = std::make_unique<SNESCorruptionInfo>();
  this->header = std::make_unique<SNESHeader>(ref(this->rom));

  // If the rom is not valid then throw an exception
  if (!this->valid())
  {
    throw InvalidSNESRomException();
  }

  // Rom is valid, set up variables.
}

void SNESCorruption::corrupt()
{
  std::cout << "Rom size: " << std::hex << rom.size() << std::dec << std::endl;

  uint32_t corruptions = 0;
  std::uniform_int_distribution<int> dist(0x00, 0xFF);

  for (uint32_t i = this->header->begin() + info->start;    // Start at header + start value
       i <= this->rom.size() - info->step && i < info->end; // While in a valid range
       i += info->step)                                     // Increase by step size
  {
    if (info->type == CorruptionType::Shift)
    {
      //  If it's okay to put the other byte in this position then change it
      if (i + info->value < rom.size() && valid_byte(this->rom[i + info->value], i))
      {
        corruptions++;
        this->rom[i] = this->rom[i + info->value];
      }
    }
    else if (info->type == CorruptionType::Swap)
    {
      
      if (i + info->value < rom.size() &&
          valid_byte(this->rom[i + info->value], i) &&  //  If it's okay to put the other byte in this position
          valid_byte(this->rom[i], i + info->value))    //  And it's okay to put this byte in the other position
      {
        corruptions++;
        uint8_t temp = this->rom[i + info->value];
        this->rom[i] = temp;
        this->rom[i + info->value] = temp;
      }
    }
    else if (info->type == CorruptionType::Add)
    {
      //  If the new byte value is valid then do the corruption
      if (valid_byte(this->rom[i] + info->value, i))
      {
        corruptions++;
        this->rom[i] += info->value;
      }
    }
    else if (info->type == CorruptionType::Set)
    {
      //  If the set value can be placed here then do it
      if (valid_byte(info->value, i))
      {
        corruptions++;
        this->rom[i] = info->value;
      }
    }
    else if (info->type == CorruptionType::Random)
    {
      bool corrupted = false;

      //  Try up to 100 times to corrupt
      for (uint32_t retry = 0; !corrupted && retry < 100; retry++)
      {
        uint8_t rand = dist(this->random);
        if (valid_byte(rand, i))
        {
          corruptions++;
          this->rom[i] = rand;
          corrupted = true;
        }
      }
    }
    else if (info->type == CorruptionType::RotateLeft)
    {
      uint8_t rotate = Util::rol<uint8_t>(this->rom[i], info->value);

      if (valid_byte(rotate, i))
      {
        this->rom[i] = rotate;
        corruptions++;
      }
    }
    else if (info->type == CorruptionType::RotateRight)
    {
      uint8_t rotate = Util::ror<uint8_t>(this->rom[i], info->value);

      if (valid_byte(rotate, i))
      {
        this->rom[i] = rotate;
        corruptions++;
      }
    }
    else if (info->type == CorruptionType::LogicalAnd)
    {
      if (valid_byte(this->rom[i] & info->value, i))
      {
        this->rom[i] &= info->value;
        corruptions++;
      }
    }
    else if (info->type == CorruptionType::LogicalOr)
    {
      if (valid_byte(this->rom[i] | info->value, i))
      {
        this->rom[i] |= info->value;
        corruptions++;
      }
    }
    else if (info->type == CorruptionType::LogicalXor)
    {
      if (valid_byte(this->rom[i] ^ info->value, i))
      {
        this->rom[i] ^= info->value;
        corruptions++;
      }
    }
    else if (info->type == CorruptionType::LogicalComplement)
    {
      if (valid_byte(~this->rom[i], i))
      {
        this->rom[i] = ~this->rom[i];
        corruptions++;
      }
    }
    else
    {
      break;  //  No corruption selected, might as well quit.
    }
  }
  std::cout << "Replaced a total of " << corruptions << " bytes." << std::endl;
}

/*
  Determines whether or not a given byte is valid to include into the rom.

  @param byte - The byte that needs to be checked
  @param location - The offset that the byte will be placed at in the rom

  @return true if the byte can be added at @location in the rom.
*/
bool SNESCorruption::valid_byte(uint8_t byte, uint32_t location)
{
  static std::vector<uint8_t> invalid_write{ 
  // BRK,  COP,  STP,  WAI,  REP,  SEP
    0x00, 0x02, 0xdb, 0xcb, 0xc2, 0xe2              // Miscellaneous invalid bytes
  };

  static std::vector<uint8_t> invalid_stack{
    0x08, 0x0B, 0x28, 0x2b, 0x48, 0x4b, 0x5a, 0x62, // Stack opcodes
    0x68, 0x7a, 0x8b, 0xab, 0xd4, 0xda, 0xf4,       // Stack opcodes
  };

  //  Map of jump/branch opcodes and their parameter size excluding the opcode byte
  //  std::map<opcode, param_size>
  static std::map<uint8_t, uint8_t> invalid_branch{
    { 0x80, 2 }, { 0x82, 3 }, { 0x30, 2 }, { 0x50, 2 }, { 0x70, 2 }, { 0x90, 2 }, { 0xb0, 2 },
    { 0xd0, 2 }, { 0xf0, 2 }, { 0x10, 2 }, { 0x4c, 3 }, { 0x6c, 3 }, { 0xdc, 3 }, { 0x7c, 3 },
    { 0x5c, 4 }, { 0x20, 3 }, { 0xfc, 3 }, { 0x22, 4 }, { 0x40, 1 }, { 0x60, 1 }, { 0x6b, 1 } 
  };

  //  STA $xxxx (, *) opcodes
  static std::vector<uint8_t> sta_ops{ 0x8D, 0x99, 0x9D };

  //  If the location points to the SNES header then it is invalid
  if (location >= this->header->offset() && location <= this->header->offset() + SNESHeader::Header_Size)
  {
    return false;
  }

  if (std::find(invalid_write.begin(), invalid_write.end(), byte) != invalid_write.end())
  {
    //  Given byte value is invalid for any space
    return false;
  }

  //  Check special case of REP (0xc2) and SEP (0xe2) opcodes which cannot
  //  have their opcode or parameter modified safely.
  if (this->rom[location - 1] == 0xc2 || this->rom[location - 1] == 0xe2)
  {
    return false;
  }

  if (std::find(invalid_stack.begin(), invalid_stack.end(), this->rom[location]) != invalid_stack.end())
  {
    //  Byte at overwrite location contains an important value.
    return false;
  }

  //  Iterate over each jump/branch instruction pair
  for (auto& pair : invalid_branch)
  {
    //  For readabiity, assign branch pair values to locals
    uint8_t opcode = pair.first;  //  Branch opcode
    uint8_t param = pair.second;  //  Branch parameter size

    for (int i = 0; i < param; i++)
    {
      //  If this part of rom contains a branch opcode return false
      if (this->rom[location - i] == opcode)
      {
        return false;
      }
    }
  }

  //  Check for STA commands
  for (auto& sta : sta_ops)
  {
    //  All 3 are 3 bytes long, so check back 2 bytes max
    for (int i = 0; i < 3; i++)
    {
      //  Check if given location has a STA opcode
      if (this->rom[location - i] == sta)
      {
        //  If opcode is found, determine whether or not it contains an important register value
        if (SNESCorruption::is_register(Util::read<uint16_t>(ref(rom), location - i)))
        {
          //  If it contains an important register then it isn't safe to overwrite.
          return false;
        }
      }
    }
  }

  return true;
}

/*
  Prints out basic header information for debugging purposes.
*/
void SNESCorruption::print_header()
{
  std::cout << std::hex;

  std::cout << "Name:     " << this->header->name() << std::endl;
  std::cout << "Layout:   " << static_cast<int>(this->header->layout())       << std::endl;
  std::cout << "Type:     " << static_cast<int>(this->header->type())         << std::endl;
  std::cout << "Rom Size: " << static_cast<int>(this->header->rom_size())     << std::endl;
  std::cout << "Ram Size: " << static_cast<int>(this->header->ram_size())     << std::endl;
  std::cout << "Country:  " << static_cast<int>(this->header->country())      << std::endl;
  std::cout << "License:  " << static_cast<int>(this->header->license_code()) << std::endl;
  std::cout << "Version:  " << static_cast<int>(this->header->version())      << std::endl;

  std::cout << std::dec;
}

/*
  Determines whether or not the currently loaded rom
  is a valid SNES rom by checking if the header exists.

  @return true for a valid header or false otherwise.
*/
bool SNESCorruption::valid()
{
  return this->header->offset() != 0;
}

void SNESCorruption::parse(std::vector<std::string> args)
{
  std::istringstream iss;

  info = std::make_unique<SNESCorruptionInfo>();

  for (uint32_t i = 0; i < args.size(); i++)
  {
    auto& arg = args[i];

    if (arg == "-s" || arg == "--step")
    {
      info->step = Util::to_int32(args[++i]);
    }
    else if (arg == "-h" || arg == "--shift")
    {
      info->value = info->value = Util::to_int32(args[++i]);
      info->type = CorruptionType::Shift;
    }
    else if (arg == "-w" || arg == "--swap")
    {
      info->value = info->value = Util::to_int32(args[++i]);
      info->type = CorruptionType::Swap;
    }
    else if (arg == "-a" || arg == "--add")
    {
      info->value = Util::to_int32(args[++i]);
      info->type = CorruptionType::Add;
    }
    else if (arg == "-t" || arg == "--set")
    {
      info->value = Util::to_int32(args[++i]);
      info->type = CorruptionType::Set;
    }
    else if (arg == "-r" || arg == "--random")
    {
      info->type = CorruptionType::Random;
    }
    else if (arg == "--rotate-left")
    {
      info->value = Util::to_int32(args[++i]);
      info->type = CorruptionType::RotateLeft;
    }
    else if (arg == "--rotate-right")
    {
      info->value = Util::to_int32(args[++i]);
      info->type = CorruptionType::RotateRight;
    }
    else if (arg == "--logical-and")
    {
      info->value = Util::to_int32(args[++i]);
      info->type = CorruptionType::LogicalAnd;
    }
    else if (arg == "--logical-or")
    {
      info->value = Util::to_int32(args[++i]);
      info->type = CorruptionType::LogicalOr;
    }
    else if (arg == "--logical-xor")
    {
      info->value = Util::to_int32(args[++i]);
      info->type = CorruptionType::LogicalXor;
    }
    else if (arg == "--logical-complement")
    {
      info->type = CorruptionType::LogicalComplement;
    }
    else if (arg == "-b" || arg == "--start")
    {
      info->start = Util::to_int32(args[++i]);
    }
    else if (arg == "-e" || arg == "--stop")
    {
      info->end = Util::to_int32(args[++i]);
    }
    else if (arg == "-o" || arg == "--out")
    {
      info->outfile = args[i + 1];
      i++;
    }
    else
    {
      throw InvalidArgumentException(arg + " is not a recognized flag.");
    }
  }
}

/*
  Saves the current ROM to a file, automatically taking into account
  the extension.

  @param filename - The file to save to. Does not require extension.
*/
void SNESCorruption::save(std::string filename)
{
  std::string format = ".sfc";

  //  If the filename doesn't include the format extension then add it.
  if (filename.length() <= 4 || filename.compare(filename.length() - format.length(), format.length(), format) != 0)
  {
    filename += format;
  }

  try
  {
    //  If the file had an SMC header then delete it.
    int offset = this->header->smc_header() ? 0 : 512;

    if (info->outfile != "")
    {
      this->rom_file.open(info->outfile, std::ios::out | std::ios::binary);
    }
    else
    {
      this->rom_file.open(filename, std::ios::out | std::ios::binary);
    }

    std::copy(rom.begin(), rom.end(), std::ostream_iterator<uint8_t>(this->rom_file));
    this->rom_file.close();

    this->save_name = filename;
  }
  catch (InvalidFileNameException e)
  {
    std::cout << e.what() << std::endl;
    exit(EXIT_FAILURE);
  }
}
